// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
  Abstract interface definition for DAL

  (c) 1997 European Space Agency - written by Marco Beijersbergen, CRI
                                              Mark Thomas, CARA

  $Id: Dal.h,v 1.3 2012/07/17 16:19:56 rsaxton Exp $

*/

#ifndef DAL_H
#define DAL_H

#include <string>
#include <vector>
#include <iosfwd>
using std::ostream;

#include "Child.h"

#include <iosfwd>


// Forward declarations.
class Attribute;
class Attributable;
class DataSet;
class Block;
class Table;
class Column;
class Array;
class ColumnData;
class CellData;
class ArrayData;
class DataSetReaderWriter;

#include "machine_specific_types.h"

/**
 * @file   Dal.h
 * 
 * @brief  The file Dal.h contains the C++ interface to the XMM-Newton
 * Data Access Layer, part of the Science Analysis System.
 * 
 * 
 */


typedef std::vector<unsigned long> longVec;


/**
 * Properties common to all DAL objects.
 * 
 */

class Dal
{
public:
    virtual ~Dal() {}
  
/**
 *  Types for columns and arrays.
 * 
 */

    enum DataType {
	Bool = 1, /**< Boolean */
	Int8, /**< 8-bit unsigned integers */
	Uint16, /**< 16-bit unsigned integers */
	Int16, /**< 16-bit signed integers */
	Uint32, /**< 32-bit unsigned integers */
	Int32, /**< 32-bit signed integers */
	Real32, /**< 32-bit reals */
	Real64, /**< 64-bit reals */
	DString /**< Strings of characters */
    };

    /** The AccessMode determines whether the data is read upon open
     * and written upon close. */

    enum AccessMode {
	Read = 1, /**< Read-only, it cannot be modified. */
	Create, /**< The data will be created. If it already exists it
		   will be overwritten. */
	Modify, /**< The data already exists, and can be modified. */
	Temp, /**< The data is created in a temporary file that is
		 discarded upon closure. */
	AsParent /**< Use the same access mode as the hierarchical
		    parent. */
    };

    /** Value to be used to indicate scalar dimension. */
    static const longVec scalar;

    /** Value to be used to indicate a starting position or a zero length. */
    static const longVec zero;

    /** Get the access mode of the object. */
    virtual AccessMode mode() const = 0;

    /**
     * Release the object. What happens to a released object may
     depend on the  memory model in use.
     * 
     */

    virtual void release() = 0;
};


/** The DataServer is an abstract interface for an object that keeps
 * track of opening and closing data files.
 */ 
class DataSetServer 
{
public :

    /** Memory Model type.
     */
    enum MemoryModel {
	High = 1, /**< Upon opening a dataset all of it is read in
		     memory. */
	HighLow, /**< Upong opening a dataset, only its structure is
		    read in memory. Data are further read as
		    required. */
	Low, /**< @see HighLow */
	UseEnvironment /**< Select a memory model based on the
			  environment variable SAS_MEMORY_MODEL. If
			  the variable is not set this is at the
			  moment the same as specifying High. */
    };


/** @name Client methods

Different clients can interact with the DataSetServer. The client name
must be set before a dataset is opened.

@param clientName: the name of the client.

*/

/*@{*/

    /** Set the name of the client. */
    
    virtual void client(const std::string& clientName) = 0;
    
    /** Retrieve the name of the current client. */
    virtual const std::string& client() const = 0;

    /** Flush all the datasets associated with the named client back to disk. */
    virtual void cleanupClient( const std::string& clientName ) = 0;


/*@}*/

/** @name Process methods

A process is a more detailed description of what the client is doing.

	@param processDescription The name of the process.
	@param dataSet: the dataset affected.

*/

/*@{*/

    /** Register a description of the process that is going on.  When
        a dataset is closed processDescription is added to the process
        list.
    */
    virtual void process( const std::string& processDescription ) = 0;

/**
 * Retrieve the name of the current process.
 * 
 */

    virtual const std::string& process() const = 0;

    /** Add processDescription to the given dataset's process
        history list.
    */
    virtual void process(
	const DataSet*		dataSet,
	const std::string&		processDescription ) = 0;

/**
 * Retrieve a process description by number.
*/

    virtual const std::string& process(
	const DataSet*		dataSet,
	unsigned int		processNumber ) const = 0;
/**
 * Return the number of procee description records associated with a
 dataset.
 * 
 */

    virtual unsigned int processes( const DataSet* dataSet ) const = 0;

/*@}*/


/** @name Debug methods
 * 
 */


/*@{*/

    /** This adds debugDescription to the given dataset's debug
        history list.
    */
    virtual void debug(
	const DataSet*		dataSet,
	const std::string&		debugDescription ) = 0;

    /**
     * Retrieve a debug description record by number.
     * 
     */

    virtual const std::string& debug(
	const DataSet*		dataSet,
	unsigned int		debugNumber ) const = 0;

/**
 * Returns how many debug description records the dataset contains.
 * 
 */

    virtual unsigned int debugs( const DataSet* dataSet ) const = 0;

/*@}*/

/**
 * @name Date and creator.
 * 
 */

/*@{*/
    /** Set the dataset creator.
     */
    virtual void creator(
	const DataSet*		dataSet,
	const std::string&		creator ) = 0;

/**
 * Retrieve the dataset creator.
 * 
 */


    virtual const std::string& creator( const DataSet*	dataSet ) const = 0;

    /**
     * Set the creation data for a dataset.
     * 
     */


    virtual void date(
	const DataSet*		dataSet,
	const std::string		date ) = 0;


    /**
     * Retrieve the creation data for a dataset.
     * 
     */


    virtual const std::string& date( const DataSet*	dataSet ) const = 0;

/*@}*/


/**
 * @name Open and close
 * 
 */


/*@{*/
    /** Open a dataset.  */
    virtual DataSet * open(
	const std::string &		setName,
	Dal::AccessMode		openMode,
	MemoryModel		memoryModel = UseEnvironment,
	DataSetReaderWriter*	readerWriter = 0 ) = 0;
    
/**
 * Make a copy of a dataset and open it.
 * 
 */

    virtual DataSet * clone(
	const std::string &		from,
	const std::string &		to,
	Dal::AccessMode		openMode,
	MemoryModel		memoryModel = UseEnvironment,
	DataSetReaderWriter*	readerWriter = 0 ) = 0;


    /**
     * Close a dataset.
     * 
     */

    virtual void close( DataSet * dataSet ) = 0;

/*@}*/

/**
 * @name Persistence.

 Datasets can be kept in memory even after the have been closed.
 * 
 */

/*@{*/

    /** Register the sets to be kept in memory or discarded.
     */
    virtual void keep( const std::string&	setName ) = 0;
    virtual void discard( const std::string&	setName ) = 0;

/*@}*/


/**
 * @name Inquiry
 * 
 */

/*@{*/

    virtual bool exists( const std::string & setName ) const = 0;
    virtual void remove( const std::string& setName ) = 0;
    virtual void copy( const std::string& from, const std::string& to ) = 0;

/*@}*/


/**
 * @name Clobber
 * 
 */

/*@{*/

    virtual void clobberDataSet( const std::string& setName ) = 0;
    virtual void clobber(bool b) = 0;
    virtual bool clobber() const = 0;
/*@}*/


    virtual void printOn(std::ostream& os) const = 0;

};

inline std::ostream& operator<<(std::ostream& os, const DataSetServer& d)
{ d.printOn(os); return os; }

// Single global instance of a DataSetServer
extern DataSetServer* dataSetServer; 


/** An object which is derived from Labelled will have a name, and have
 * an associated short textual description (which is essentially
 * a comment).
 * The name of the object is obtained with name()
 * The short textual description is obtained with label()
 * The object may be renamed with rename() and
 * the short textual description may be changed (i.e. replaced)
 * with relabel().
 * qualifiedName() returns a colon-separated concatenation of the
 * names, in hierarchical order, or the object and all its
 * ancesters, up to and including the data set name.
 */
class Labelled
{
public :
    virtual ~Labelled(){}
    virtual const std::string&	name() const = 0;
    virtual void		rename( const std::string & newName ) = 0;
    virtual const std::string&	label() const = 0;
    virtual void		relabel( const std::string & newLabel ) = 0;
    virtual std::string		qualifiedName() const = 0;
};


/** An object that is derived from Attributable has a set of attributes.
 * An attribute is a dictionary of keyword-value pairs. Numeric attributes
 * have a string describing the units; each attribute has a comment.
 */
class Attributable : public virtual Dal, public virtual Labelled
{
public:

    virtual ~Attributable() {}

    virtual Attributable& operator=( const Attributable& ) = 0;

    virtual Attribute * addAttribute(
	const std::string &		name,
	const std::string &		comment = "",
	const std::string &		units = "" ) = 0;

    virtual Attribute * addAttribute( const Attribute * attribute ) = 0;

    virtual void addAttributes( const Attributable* ) = 0;

    /** Returns true if an attribute of the specified name exists. */
    virtual bool hasAttribute( const std::string & attributeName ) const = 0;

    /** Returns the attribute of the given name. If it does not exist, an
        error is generated. */
    virtual const Attribute * attribute(
	const std::string&	attributeName,
	AccessMode	accessMode=AsParent ) const = 0;
    virtual Attribute * attribute(
	const std::string&	attributeName,
	AccessMode	accessMode=AsParent ) = 0;

    /** Returns the number of attributes. */
    virtual unsigned int attributes() const = 0;

    /** Returns the attribute of the given number. */
    virtual const Attribute * attribute(
	unsigned int	number,
	AccessMode	accessMode=AsParent ) const = 0;
    virtual Attribute * attribute(
	unsigned int	number,
	AccessMode	accessMode=AsParent ) = 0;

    /** Returns the number of the given attribute. */
    virtual unsigned int attributeNumber(
	const std::string&	attributeName ) const = 0;

    /** Deletes the attribute with the given name. */
    virtual void deleteAttribute( const std::string & name ) = 0;

    /** Deletes attribute with the given number. */
    virtual void deleteAttribute( unsigned int number ) = 0;

    virtual void addComment(const std::string& comment) = 0;

    /** Returns the number of comment lines. */
    virtual unsigned int comments() const = 0;

    /** Returns comment line with the specified number starting from 0. */
    virtual const std::string & comment( unsigned int number ) const = 0;

    virtual void addHistory(const std::string& historyComment) = 0;

    /** Returns the number of history records. */
    virtual unsigned int historys() const = 0;

    /** Returns the specified history record (starting from 0). */
    virtual const std::string & history ( unsigned int number ) const = 0;

    virtual void printOn(std::ostream& os) const = 0;
};

inline std::ostream& operator<<(std::ostream& os, const Attributable& d)
{ d.printOn(os); return os; }


class Attribute : public virtual Dal, public virtual Labelled, public virtual Child<Attributable>
{
public :

    enum DataType { Bool=1, Int, Real, DString };

    virtual ~Attribute() {}

    // explicitly add char* since the egcs compiler will convert
    // the char* to bool, not to string.
    virtual Attribute&	operator=( int	value ) = 0;
    virtual Attribute&	operator=( double	value ) = 0;
    virtual Attribute&	operator=( bool	value ) = 0;
    virtual Attribute&	operator=( const char * value ) = 0;
    virtual Attribute&	operator=( const std::string & value ) = 0;
    virtual Attribute&	operator=( const Attribute& ) = 0;

    /** Returns the value of the attribute, converting it if necessary.
     * An error is generated in case it is not possible to convert
     * value to the desired type. */
    virtual int		asInt() const = 0;
    virtual double		asReal() const = 0;
    virtual std::string		asString() const = 0;
    virtual bool		asBool() const = 0;

    virtual const std::string&	units() const = 0;
    virtual void		units(const std::string&) = 0;

    virtual DataType	dataType() const = 0;
    virtual void		dataType(DataType type) = 0;
    virtual char*		addressOfValue() const = 0;
    virtual unsigned int	dataSize() const = 0;
    virtual unsigned int	attributeNumber() const = 0;

    virtual void		printOn(std::ostream& os) const = 0;
};

inline std::ostream& operator<<(std::ostream& os, const Attribute& d)
{ d.printOn(os); return os; }


// Structure classes. These objects do not necessarily contain the data
// itself, but they contain the information associated with the objects.
// For example, the column object can tell you its name, data type, number
// of rows etc, but to access the data the data in the column itself
// one of the data objects is needed.  

/** A DataSet is a set of blocks, where a block is either a data table or
 * an array.
 */ 
class DataSet : public virtual Attributable, public virtual Child<DataSetServer>
{

public:

    virtual ~DataSet() {}

    virtual DataSet& operator=( const DataSet& ) = 0;

    virtual Table * addTable(
	const std::string &	name,
	unsigned long	rows,
	const std::string &	label = "",
	int		position = -1 ) = 0;

    virtual Array * addArray(
	const std::string&			name,
	DataType			dataType,
	const longVec &			size,
	const std::string &			units = "",
	const std::string &			label = "",
	int				position = -1 ) = 0;

    virtual Block * add(
	const Block *		block,
	const std::string& 		newName = "",
	int			position = -1 ) = 0;

    /** Returns true if a block of the specified name exists. */
    virtual bool		hasBlock( const std::string & blockName ) const = 0;

    virtual Table *		table( unsigned int blockNumber, AccessMode=AsParent ) = 0;
    virtual const Table *	table( unsigned int blockNumber, AccessMode=AsParent ) const = 0;
    virtual Table *		table( const std::string & blockName, AccessMode=AsParent ) = 0;
    virtual const Table *	table( const std::string & blockName, AccessMode=AsParent ) const = 0;
    virtual Array *		array( unsigned int blockNumber, AccessMode=AsParent ) = 0;
    virtual const Array *	array( unsigned int blockNumber, AccessMode=AsParent ) const = 0;
    virtual Array *		array( const std::string & blockName, AccessMode=AsParent ) = 0;
    virtual const Array *	array( const std::string & blockName, AccessMode=AsParent ) const = 0;
    virtual Block *		block( unsigned int blockNumber, AccessMode=AsParent ) = 0;
    virtual const Block *	block( unsigned int blockNumber, AccessMode=AsParent ) const = 0;
    virtual Block *		block( const std::string & blockName, AccessMode=AsParent ) = 0;
    virtual const Block *	block( const std::string & blockName, AccessMode=AsParent ) const = 0;
    virtual void		deleteBlock(unsigned blockNumber) = 0;
    virtual void		deleteBlock( const std::string & blockName ) = 0;

    /** Returns the number of the block of the given name. */
    virtual unsigned		blockNumber(const std::string & name) const = 0;

    /** Returns the number of blocks in the data set. */
    virtual unsigned		blocks() const = 0;

    //virtual void		printOn(std::ostream& os) const = 0;
};

//inline std::ostream& operator<<(std::ostream& os, const DataSet& d)
//{ d.printOn(os); return os; }


/** A block is an abstract interface for all component of a DataSet. */
class Block : public virtual Attributable, public virtual Child<DataSet>
{

 public:

    enum Type { TableT = 1, ArrayT };

    virtual ~Block() {}

    virtual Type type() const = 0;
    virtual unsigned int blockNumber() const = 0;
    //virtual void printOn(std::ostream&) const = 0;

};

//inline std::ostream& operator<<(std::ostream& os, const Block& b)
//{ b.printOn(os); return os; }


template<class T>
class Seekable
{
public :

    virtual ~Seekable() {}
    virtual void seek( T from, T count ) = 0 ;
    virtual T from() const = 0;
    virtual T count() const = 0;

};


class Table : public virtual Block, public virtual Seekable<unsigned long>
{

public:

    virtual ~Table() {}

    virtual Column * addColumn(
	const std::string&			name,
	DataType			dataType,
	const std::string&			label = "",
	const std::string&			units = "",
	const longVec & 		size = scalar,
	int				position = -1 ) = 0;

    virtual Column * add(
	const Column *	column,
	const std::string& 	newName = "",
	int		position = -1 ) = 0;

    virtual Column * column(unsigned columnNumber, AccessMode=AsParent ) = 0;
    virtual const Column * column(unsigned columnNumber, AccessMode=AsParent ) const = 0;
    virtual Column * column(const std::string & columnName, AccessMode=AsParent ) = 0;
    virtual const Column * column(const std::string & columnName, AccessMode=AsParent ) const = 0;
    virtual bool hasColumn( const std::string & columnName ) const = 0;
    virtual unsigned int columnNumber(const std::string & columnName ) const = 0;
    virtual void deleteColumn(unsigned columnNumber ) = 0;
    virtual void deleteColumn( const std::string & columnName ) = 0;
    virtual unsigned long rows() const = 0;
    virtual unsigned columns() const = 0;

    virtual void copyRows( unsigned long from, unsigned long to, unsigned long count=1 ) = 0;
    virtual void deleteRows( unsigned long from, unsigned long count=1 ) = 0;
    virtual void insertRows( unsigned long pos, unsigned long count=1 ) = 0;

    /** Writes an ascii representation of the column to a stream. */
    //virtual void printOn(std::ostream& os) const = 0;
    virtual void forEachSubTable( void (*callThisFunction)(Table *) ) = 0;
};


/** Nullable allows the values in a data component to have a designated
 * null (or undefined) value.
 */
class Nullable
{
public :

    virtual ~Nullable() {};

    /** Null Value type */
    enum NullType { Integer = 1, Real, String, Undefined };
    virtual NullType nullType() const = 0;

    /** Set/Get Null value */
    virtual void nullValue( long value ) = 0;
    virtual long intNullValue() const = 0;
    virtual bool nullDefined() const = 0;

    /** Delete the null value (error if not defined) */
    virtual void deleteNullValue() = 0;
};


/** A DataComponent is a collection of values all of the same type,
 * arranged in a multidimensional array.
 */
class DataComponent : public virtual Dal, public virtual Nullable
{
public :

    virtual ~DataComponent() {};

    /** Basic information about the size of the matrices. */
    virtual DataType dataType() const = 0;
    virtual unsigned int dimensions() const = 0;
    virtual unsigned long elements() const = 0;
    virtual const longVec& size() const = 0;

    /** Get and set units */
    virtual const std::string & units() const = 0;
    virtual void units( const std::string& ) = 0;

    /** The size in bytes of a single value. */
    virtual unsigned int dataSize() const = 0;

    /** Scaling */
    virtual void scaling( double zero, double scale ) = 0;
    virtual void scale( bool onoff ) = 0;
    virtual bool scaled() const = 0;
    virtual double scaleZero() const = 0;
    virtual double scaleFactor() const = 0;

    /** Writes an ascii representation of the column to a stream. */
    //virtual void printOn(std::ostream& os) const = 0;
};


class Column : public virtual DataComponent, public virtual Attributable, public virtual Child<Table>, public virtual Seekable<unsigned long>
{

public:

    enum CellType { Fixed = 1, Variable };

    virtual ~Column() {}

    virtual unsigned int columnNumber() const = 0;
    virtual CellType cellType() const = 0;

    virtual unsigned long rows() const = 0;

    virtual ColumnData * data(
	unsigned long			from=0,
	unsigned long		 	count=0,
	AccessMode			accessMode=AsParent ) const = 0;

    /** Get exactly one cell; will work for either variable or fixed length. */
    virtual ColumnData * rowData(
	unsigned long			rowNumber,
	AccessMode			accessMode=AsParent ) const = 0;

    virtual CellData  * cellData(
	unsigned long			rowNumber,
	unsigned long 			from=0,
	unsigned long 	 		count=0,
	AccessMode			accessMode=AsParent ) const = 0;

	std::vector<std::string> _stringData;

    /** Writes an ascii representation of the column to a stream. */
    //virtual void printOn(std::ostream& os) const = 0;
};

//inline std::ostream& operator<<(std::ostream& os, const Column& c)
//{ c.printOn(os); return os; }


/** An array is a Block that consists of an n-dimensional array of values,
 *each value of the same type.
 */
class Array : public virtual DataComponent, public virtual Block, public virtual Seekable< longVec >
{

 public:

    virtual ~Array() {}

    /** Setup an object to access the contents of this array. */
    virtual ArrayData * data(
	const longVec&		from=zero,
	const longVec&		count=zero,
	AccessMode		accessMode=AsParent) const = 0;


};


// Data access objects.

class Data : public virtual Dal
{
public :

    virtual ~Data(){}

    /** isNull() : Determine if value is null.
     * setToNull() : Set value to null.
     * hasNulls() : Determine if object has any null values.
     */
    virtual bool isNull( unsigned long pos ) const = 0;
    virtual void setToNull( unsigned long pos ) const = 0;
    virtual bool hasNulls() const = 0;

    /** Return a pointer to the start of the data. Generates an error
     * in case the data cannot be accessed as a contiguous chunck of memory,
     * such as a variable-size column.
     */
    virtual int8*   	int8Data() const = 0;
    virtual uint16* 	uint16Data() const = 0;
    virtual int16*  	int16Data() const = 0;
    virtual uint32* 	uint32Data() const = 0;
    virtual int32*  	int32Data() const = 0;
    virtual real32* 	real32Data() const = 0;
    virtual real64* 	real64Data() const = 0;
    virtual bool8*  	boolData() const = 0;
    virtual char*   	stringData() const = 0;

    template<class T> T * data() const;

    virtual unsigned int 	dataSize() const = 0;
    virtual unsigned long 	elements() const = 0;
    virtual DataType 	dataType() const = 0;
    virtual void 		printOn(std::ostream& os, const std::string& sep="") const = 0;


};


template<> inline int8   * Data::data<int8>() const { return int8Data(); }
template<> inline int16  * Data::data<int16>() const { return int16Data(); }
template<> inline uint16 * Data::data<uint16>() const { return uint16Data(); }
template<> inline int32  * Data::data<int32>() const { return int32Data(); }
template<> inline uint32 * Data::data<uint32>() const { return uint32Data(); }
template<> inline real32 * Data::data<real32>() const { return real32Data(); }
template<> inline real64 * Data::data<real64>() const { return real64Data(); }
// Because there is a typedef char bool8 in the system, it is not possible
// to offer a specialization for both bool8 and char.
template<> inline char  * Data::data<char>() const {
    switch(dataType()){
	case Dal::DString:
	    return stringData();
	    break;
	case Dal::Bool:
	    return boolData();
	    break;
	default: {}
	    //  Sas::error("UnhandledType", "template<typename T> Data::data()");
    }
    return 0;
}


inline std::ostream& operator<<(std::ostream& os, const Data& d)
{ d.printOn(os); return os; }


/** The ColumnData object gives access to (part of) the data in a column.
 * This object is constructed by a Column, but has to be explicitly
 * deleted after it's last use.
 */
class ColumnData : public virtual Data, public virtual Seekable<unsigned long>, public virtual Child<Column>
{
public :

    virtual ~ColumnData() {}

    virtual const longVec& size() const = 0;
    virtual unsigned long position(
	unsigned long			row) const = 0;

    virtual unsigned long position(
	unsigned long			row,
	unsigned long			pos) const = 0;

    virtual unsigned long position(
	unsigned long			row,
	const longVec& 			pos) const = 0;

};

/** MatrixData provides an interface to a rectangular multi-dimensional array.
 */
class MatrixData : public virtual Data, public virtual Seekable< longVec >
{

public:

    virtual ~MatrixData() {}

    virtual unsigned long position(const longVec& pos) const = 0;

};


/** The CellData object gives access to (part of) a single cell in a column.
 * This object is constructed by a Column, but has to be explicitly
 * deleted after it's last use.
 * For Variable length columns, the number of dimensions is 1.
 */
class CellData : public virtual Data, public virtual Seekable<unsigned long>, public virtual Child<Column>
{
public :

    virtual ~CellData() {}
    virtual unsigned long size() const = 0;
    virtual void size( unsigned long size ) = 0;
    virtual unsigned long row() const = 0;

};


/** The ArrayData object gives access to (part of) the data in an Array.
 * This object is constructed by an Array, but has to be explicitly
 * deleted after it's last use.
 */
class ArrayData : public virtual MatrixData, public virtual Child<Array>
{
public :
    virtual ~ArrayData(){}
    virtual const longVec& size() const = 0;
};



#endif
