// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;
/*************************************************************** 
* 
* CompoundName.cc             & XMM Science Analysis System (SAS) 
* 
* Various CompoundName utilities
* 
* (c) 2000 ESA - Written by Mark Thomas, CARA
* 
* $Id: CompoundName.cc,v 1.2 2004/03/05 14:43:00 eojero Exp $ 
* 
*****************************************************************/ 

#include <Tracer.h>
#include <errstr.h>

#include "CompoundName.h"

CompoundName::CompoundName(
			const string&			name,
			SyntaxType			expected,
			DataSet*			dataSet,
			Dal::AccessMode 		accessMode,
			DataSetServer::MemoryModel	memoryModel ) :
				_name( name ),
				_expected( expected ),
				_accessMode( accessMode ),
				_memoryModel( memoryModel ),
				_dataSet( dataSet ),
				_set( dataSet )
				//_block( 0 ),
				//_array( 0 ),
				//_table( 0 ),
				//_column( 0 ),
				//_attribute( 0 ),
				//_attributable( 0 )
{
	PROC

	parseCompoundName();

	ENDPROC
}


void
CompoundName::parseCompoundName()
{
	PROC

	SimpleScanner	sc( _name, ":%" );

	parseColumnName( sc );

	if( sc.nextIsCharacter( '%' ) )
	{
		parseAttributeName( sc );
	} // if

	if( !sc.eof() )
	{
		errstr << "The compound name '" << _name << "' is invalid; trailing characters found." << Sas::error( "invalidCompoundName" );
	} // if

	// ***************
	// Error checking.
	// ***************
	if( _expected == CompoundName::AnyAttributeName )
	{
		if(
			_state != CompoundName::DataSetAttributeName
			&&
			_state != CompoundName::BlockAttributeName
			&&
			_state != CompoundName::ColumnAttributeName )
		{
			errstr << "The name " << _name << " was expected to be an attribute name" << Sas::error( "attributeNameExpected" );
		} // if
	} // if

	// **************************
	// Set final value of _state.
	// **************************
	if( _dataSet )
	{
		if( _expected == CompoundName::AnyDataComponentName )
		{
			if( _state == CompoundName::BlockName )
			{
				if( _dataSet -> block( _name ) -> type() == Block::ArrayT )
				{
					_state = CompoundName::ArrayName;
				}
				else
				{
					errstr << "Found table name " << _name << "but expected array name" << Sas::error( "expectedArrayName" );
				} // if
			} // if
		} // if
		if( _expected == CompoundName::Any )
		{
			if( _state == CompoundName::BlockName )
			{
				if( _dataSet -> block( _name ) -> type() == Block::ArrayT )
				{
					_state = CompoundName::ArrayName;
				}
				else
				{
					_state = CompoundName::TableName;
				} // if
			} // if
		} // if
	} // if
	ENDPROC
}


void
CompoundName::parseColumnName( SimpleScanner& sc )
{
	PROC

	parseBlockName( sc );

	if( sc.nextIsCharacter( ':' ) )
	{
		sc.skip();

		//if( !sc.nextIsCharacter( ':' ) )
		//{
		//	_state = CompoundName::BlankColumnName;
		//}
		//else
		if( !sc.nextIsCharacter( ':' ) && !sc.nextIsCharacter( '%') )
		{
			try
			{
				_columnName = sc.match( SimpleToken::String ).asString();
			}
			catch( SimpleScanner::Expected e )
			{
				errstr << "The compound name '" << _name << "' is invalid; a column name was expected." << Sas::error( "invalidCompoundName" );
			}
		} // if
		_state = CompoundName::ColumnName;
	} // if

	ENDPROC
}


void
CompoundName::parseBlockName( SimpleScanner& sc )
{
	PROC

	parseDataSetName( sc );

	if( sc.nextIsCharacter( ':' ) )
	{
		sc.skip();

		//if( !sc.nextIsCharacter( ':' ) )
		//{
		//	_state = CompoundName::BlankBlockName;
		//}
		//else
		if( !sc.nextIsCharacter( ':' ) && !sc.nextIsCharacter( '%') )
		{
			try
			{
				_blockName = sc.match( SimpleToken::String ).asString();
			}
			catch( SimpleScanner::Expected e )
			{
				errstr << "The compound name '" << _name << "' is invalid; a block name was expected." << Sas::error( "invalidCompoundName" );
			}
		} // if
		_state = CompoundName::BlockName;
	} // if

	//if( _dataSet && _dataSet -> name() != _dataSetName )
	//	} // if
	//} // if
	ENDPROC
}


void
CompoundName::parseDataSetName( SimpleScanner& sc )
{
	PROC

	try
	{
		_dataSetName = sc.match( SimpleToken::String).asString();
	}
	catch( SimpleScanner::Expected e )
	{
		errstr << "The compound name '" << _name << "' is invalid; a dataset name was expected." << Sas::error( "invalidCompoundName" );
	}

	//if( _dataSet && _dataSet -> name() != _dataSetName )
	//{
	//	errstr << "The compound name '" << name << "' is invalid; the dataset name component '" << _dataSetName << "' was expected to be " << _dataSet -> name() << Sas::error( "invalidCompoundName" );
	//} // if

	//if( syntaxType == CompoundName::DataSetName )
	//{
	//	if( !sc.eof() )
	//	{
	//		errstr << "The compound name " << name << " is invalid; a dataset name was expected but trailing characters found." << Sas::error( "invalidCompoundName" );
	//	} // if
	//} // if

	_state = CompoundName::DataSetName;

	ENDPROC
}


void
CompoundName::parseAttributeName( SimpleScanner& sc )
{
	PROC
	sc.match( '%' );
	//if( sc.eof() )
	//{
	//	_state = CompoundName::BlankAttributeName;
	//}
	//else
	if( !sc.eof() )
	{
		try
		{
			_attributeName = sc.match( SimpleToken::String ).asString();
		}
		catch( SimpleScanner::Expected e )
		{
			errstr << "The compound name '" << _name << "' is invalid; a dataset name was expected." << Sas::error( "invalidCompoundName" );
		}

		if( !sc.eof() )
		{
			errstr << "Trailing characters" << Sas::error( "invalidCompoundName" );
		} // if
	} // if

	switch( _state )
	{
		case CompoundName::DataSetName :
			_state = CompoundName::DataSetAttributeName;
			break;
		case CompoundName::BlockName :
			_state = CompoundName::BlockAttributeName;
			break;
		case CompoundName::ColumnName :
			_state = CompoundName::ColumnAttributeName;
			break;
		default :
			errstr << "Internal error when parsing compoundName '" << _name << "'" << Sas::error( "internalError" );
	} // switch

	ENDPROC
}


CompoundName::~CompoundName()
{
	PROC
	if( !_dataSet && _set )
	{
		dataSetServer -> close( _set );
	} // if
	ENDPROC
}


CompoundName::SyntaxType
CompoundName::state() const
{
	PROC
	RETURN _state;
}


//bool
//CompoundName::exists() const
//{
//}


Attributable*
CompoundName::attributable() const
{
	PROC

	Attributable*	thisAttributable( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _name, _accessMode, _memoryModel );
	} // if

	//if( _dataSet )
	//{
		switch( _state )
		{
			case CompoundName::DataSetAttributeName :
			case CompoundName::DataSetName :
				thisAttributable = _set;
				break;
			case CompoundName::BlockAttributeName :
			case CompoundName::BlockName :
				thisAttributable = _set -> block( _name );
				break;
			case CompoundName::ColumnAttributeName :
			case CompoundName::ColumnName :
				thisAttributable = _set -> table( _name ) -> column( _name );
				break;
			default :
				errstr << "CompoundName: Internal error" << Sas::error( "internalError" );
		} // switch
	//} // if
	RETURN thisAttributable;
}


Column*
CompoundName::column() const
{
	PROC

	Column*		thisColumn( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _dataSetName, _accessMode, _memoryModel );
	} // if

	thisColumn = _set -> table( _blockName ) -> column( _columnName );

	RETURN thisColumn;
}


Array*
CompoundName::array() const
{
	PROC

	Array*		thisArray( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _dataSetName, _accessMode, _memoryModel );
	} // if

	thisArray = _set -> array( _blockName );

	RETURN thisArray;
}


Table*
CompoundName::table() const
{
	PROC

	Table*		thisTable( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _dataSetName, _accessMode, _memoryModel );
	} // if

	thisTable = _set -> table( _blockName );

	RETURN thisTable;
}


Block*
CompoundName::block() const
{
	PROC

	Block*		thisBlock( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _dataSetName, _accessMode, _memoryModel );
	} // if

	thisBlock = _set -> block( _blockName );

	RETURN thisBlock;
}


DataSet*
CompoundName::dataSet() const
{
	PROC

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _dataSetName, _accessMode, _memoryModel );
	} // if

	RETURN _set;
}


Attribute*
CompoundName::attribute() const
{
	PROC

	Attribute*	thisAttribute( 0 );

	if( !_dataSet && !_set )
	{
		_set = dataSetServer -> open( _name, _accessMode, _memoryModel );
	} // if

	switch( _state )
	{
		case CompoundName::DataSetAttributeName :
			thisAttribute = _set -> attribute( _attributeName );
			break;
		case CompoundName::BlockAttributeName :
			thisAttribute = _set -> block( _blockName ) -> attribute( _attributeName );
			break;
		case CompoundName::ColumnAttributeName :
			thisAttribute = _set -> table( _blockName ) -> column( _columnName ) -> attribute( _attributeName );
			break;
		default :
			errstr << "CompoundName: Internal error" << Sas::error( "internalError" );
		} // switch
	//} // if
	RETURN thisAttribute;
}


string
CompoundName::dataSetName() const
{
	PROC
	RETURN _dataSetName;
}


string
CompoundName::blockName() const
{
	PROC
	RETURN _blockName;
}


string
CompoundName::columnName() const
{
	PROC
	RETURN _columnName;
}


string
CompoundName::attributeName() const
{
	PROC
	RETURN _attributeName;
}


string
CompoundName::qualifiedBlockName() const
{
	PROC
	RETURN _dataSetName + ":" + _blockName;
}


string
CompoundName::qualifiedColumnName() const
{
	PROC
	RETURN _dataSetName + ":" + _blockName + ":" + _columnName;
}


string
CompoundName::qualifiedAttributeName() const
{
	PROC
	RETURN _attributeName;
}


string
CompoundName::qualifiedAttributableName() const
{
	PROC
	string	n;
	switch( _state )
	{
		case CompoundName::DataSetName :
		case CompoundName::DataSetAttributeName :
			n = _dataSetName;
			break;
		case CompoundName::BlockName :
		case CompoundName::BlockAttributeName :
			n = _dataSetName + ":" + _blockName;
			break;
		case CompoundName::ColumnName :
		case CompoundName::ColumnAttributeName :
			n = _dataSetName + ":" + _blockName + ":" + _columnName;
			break;
		default :
			errstr << "qualified attributable name could not be determined" << Sas::error( "internalError ");
			break;
	} // switch
	RETURN n;
}



