// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;
#include<cassert>


#include <Dal.h>
#include <iostream>
#include <ctime>
#include "StopWatch.h"



static void
fill( Table * table )
{
	const unsigned long nRows = table -> rows();

	for( unsigned int i = 0; i < table -> columns(); ++i )
	{
		Column * col = table -> column( i );
		ColumnData * coldat = col -> data();

		switch( col -> dataType() )
		{
			case Column::Bool :
			{
				bool8 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> boolData();
					d[ 0 ] = ((coldat -> from()) % 2 ? true : false );
				} // for
			}
				break;
			case Column::Int8 :
			{
				int8 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int8Data();
					d[ 0 ] = coldat -> from();
				} // for
			}
				break;
			case Column::Int16 :
			{
				int16 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int16Data();
					d[ 0 ] = coldat -> from();
				} // for
			}
				break;
			case Column::Int32 :
			{
				int32 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int32Data();
					d[ 0 ] = coldat -> from();
				} // for
			}
				break;
			case Column::Real32 :
			{
				real32 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> real32Data();
					d[ 0 ] = coldat -> from();
				} // for
			}
				break;
			case Column::Real64 :
			{
				real64 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> real64Data();
					d[ 0 ] = coldat -> from();
				} // for
			}
				break;
			case Column::DString :
			{
				char * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> stringData();
					const unsigned long	e = coldat -> parent() -> elements();
					for( unsigned long k = 0; k < e; ++k )
					{
						d[ k ] = ('a' + ( coldat -> from() ) % 26 );
					} // for
				} // for
			}
				break;
			default :
				break;
		} // switch
		delete coldat;
	} // for
}


static void
check( Table * table )
{
	const unsigned long nRows = table -> rows();

	for( unsigned int i = 0; i < table -> columns(); ++i )
	{
		Column * col = table -> column( i );
		ColumnData * coldat = col -> data();

		switch( col -> dataType() )
		{
			case Column::Bool :
			{
				bool8 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> boolData();
					assert( d[ 0 ] == ((coldat -> from() ) % 2 ? true : false ));
				} // for
			}
				break;
			case Column::Int8 :
			{
				int8 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int8Data();
					assert( d[ 0 ] == int8(coldat -> from()) );
				} // for
			}
				break;
			case Column::Int16 :
			{
				int16 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int16Data();
					assert( d[ 0 ] == int16(coldat -> from()) );
				} // for
			}
				break;
			case Column::Int32 :
			{
				int32 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> int32Data();
					assert( d[ 0 ] == int32(coldat -> from()) );
				} // for
			}
				break;
			case Column::Real32 :
			{
				real32 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> real32Data();
					assert( d[ 0 ] == coldat -> from());
				} // for
			}
				break;
			case Column::Real64 :
			{
				real64 * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> real64Data();
					assert( d[ 0 ] == coldat -> from());
				} // for
			}
				break;
			case Column::DString :
			{
				char * d;
				for( unsigned long j = 0; j < nRows; ++j )
				{
					coldat -> seek( j, 1 );
					d = coldat -> stringData();
					const unsigned long	e = coldat -> parent() -> elements();
					for( unsigned long k = 0; k < e; ++k )
					{
						assert( d[ k ] = ('a' + ( coldat -> from() ) % 26 ) );
					} // for
				} // for
			}
				break;
			default :
				break;
		} // switch
		delete coldat;
	} // for
}


static void
testSeek( Table * table )
{
	for( unsigned int i = 0; i < table -> columns(); ++i )
	{
		Column * col = table -> column( i );
		ColumnData * coldat = col -> data();
		coldat -> seek( 1, 2 );

		switch( col -> dataType() )
		{
			case Column::Bool :
			{
				bool8 * d = coldat -> boolData();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					//assert( d[ j ] == ((coldat -> from() + j )) % 2 ? true : false );
					assert( d[ j ] == ((coldat -> from() + j ) % 2 ? true : false ));
				} // for
			}
				break;
			case Column::Int8 :
			{
				int8 * d = coldat -> int8Data();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					assert( d[ j ] == int8(coldat -> from() + j ) );
				} // for
			}
				break;
			case Column::Int16 :
			{
				int16 * d = coldat -> int16Data();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					assert( d[ j ] == int16(coldat ->
					from() + j) );
				} // for
			}
				break;
			case Column::Int32 :
			{
				int32 * d = coldat -> int32Data();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					assert( d[ j ] == int32(coldat ->
					from() + j) );
				} // for
			}
				break;
			case Column::Real32 :
			{
				real32 * d = coldat -> real32Data();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					assert( d[ j ] == coldat -> from() + j );
				} // for
			}
				break;
			case Column::Real64 :
			{
				real64 * d = coldat -> real64Data();
				for( unsigned long j = 0; j < coldat -> count(); ++j )
				{
					assert( d[ j ] == coldat -> from() + j );
				} // for
			}
				break;
			case Column::DString :
			{
				char * d = coldat -> stringData();
				for( unsigned long r = 0; r < coldat -> count(); ++r )
				{
					const unsigned long	e = coldat -> parent() -> elements();
					for( unsigned long j = 0; j < e; ++j )
					{
						assert( d[ r * e + j ] = ('a' + ( coldat -> from() + j ) % 26 ));
					} // for
				} // for
			}
				break;
			default :
				break;
		} // switch
		delete coldat;
	} // for
}


int main()
{
	// *********************************
	// Get clock for timing information.
	// *********************************
	StopWatch		cpuWatch( StopWatch::CPU );
	StopWatch		calenderWatch( StopWatch::Calender );

  	unsigned long nRows = 3000;

	// Create, fill & check
  	DataSet* set = dataSetServer->open("test.dat",DataSet::Create);
  	Table* tab = set->addTable("some table",nRows);
 	Column* col1; col1 = tab->addColumn("bool",Column::Bool,"bool data");
 	Column* col2; col2 = tab->addColumn("int8",Column::Int8,"int8 data");
 	Column* col3; col3 = tab->addColumn("int16",Column::Int16,"int16 data");
 	Column* col4; col4 = tab->addColumn("int32",Column::Int32,"int32 data");
 	Column* col5; col5 = tab->addColumn("real32",Column::Real32,"real32 data");
 	Column* col6; col6 = tab->addColumn("real64",Column::Real64,"Real64 data");
	vector<unsigned long>	size;
	size.push_back( 500 );
 	Column* col7; col7 = tab->addColumn("string",Column::DString,"string data", "", size );

  	fill( tab );
  	check( tab );
	dataSetServer->close(set);

	// Read & check
  	set = dataSetServer->open("test.dat",DataSet::Read);
  	tab = set->table("some table");
  	check( tab );
  	dataSetServer->close(set);

  	set = dataSetServer->open("test.dat",DataSet::Read);
  	tab = set->table("some table");
	testSeek( tab );
  	dataSetServer->close(set);

	cout << "CPU time: " << cpuWatch.elapsed() / 1000000.0 << " secs" << endl;
	cout << "Calender time: " << calenderWatch.elapsed() << " secs" << endl;

  	return 0;
}

